(** This module generates coalgebraic formulae and queries, and runs solvers on them.
    The generated formulae contain randomly generated formulae
    as well as instances of certain classes of formulae.
    @author Florian Widmann
 *)


module G = GenAndComp
module C = CoAlgFormula
module A = ALCFormula

open CoolUtils

type solvanswer = ASAT | AUNSAT | ATIMEOUT | AFAILED

exception UserInterupt

let string_of_solvanswer a =
    match a with
    | ASAT -> "sat"
    | AUNSAT -> "unsat"
    | ATIMEOUT -> "timeout"
    | AFAILED -> "failed"

type testresults =
   ((string list) (* solver names *)
    * ((string * ((solvanswer * float) list)) (*some test case with times for each reasoner*)
        list))

let debugMsg str =
    prerr_endline (":: " ^ str)

let ptimeout = ref (300) (* default timeout in seconds *)


(* generates the int range from i to j (including) *)
let (--) i j = 
    let rec aux n acc =
      if n < i then acc else aux (n-1) (n :: acc)
    in aux j []

let _ = Gc.set { (Gc.get()) with Gc.minor_heap_size = 4194304; major_heap_increment = 67108864; space_overhead = 120 }


(*
let seed =
  Random.self_init (); 
  Random.int ((1 lsl 30) - 1)
*)

let pseed = ref (209314489)

let sortK = Array.make 1 (CoAlgMisc.MultiModalK, [0])

let nomTable name =
  assert (CoAlgFormula.isNominal name);
  Some 0


(*
let maxCard = ref 100

let const_min r f =
  let card = (Random.int !maxCard) + 1 in
  C.const_min card r f

let const_max r f =
  let card = (Random.int (!maxCard+1)) in
  C.const_max card r f
 *)

let lFF = [(C.const_not, 1)]
let lFFF = [(C.const_or, 1); (C.const_and, 1)]
let lPFF = [(C.const_ex, 1); (C.const_ax, 1)]

let genFormula litFkt pFkt size =
  let nrLitsF = litFkt size in
  let lF =
    G.genLiterals ~litlst:[C.const_true; C.const_false] ~prio:10 C.const_ap "p" nrLitsF
  in
  let nrLitsP = pFkt size in
  let lP =
    if nrLitsP > 1 then G.genLiterals (fun x -> x) "a" nrLitsP
    else [("", 1)]
  in
  let genRF = fst (G.mk_generator lF lFF lFFF lPFF lP [] [] []) in
  genRF size

let genQuery litFkt pFkt tbcardFkt tbfsizeFkt size =
  let tbcard = tbcardFkt size in
  let tbfsize = tbfsizeFkt size in
  let tblitFkt n = litFkt size in
  let tbpFkt n = pFkt size in
  let tbox = ref [] in
  for i = 1 to tbcard do
    let tbf = genFormula tblitFkt tbpFkt tbfsize in
    tbox := (0, tbf)::!tbox
  done;
  let f = genFormula litFkt pFkt size in
  (!tbox, (0, f))


let rec exportALCFormula f =
  match f with
  | C.TRUE -> A.TRUE
  | C.FALSE -> A.FALSE
  | C.AP s when not (C.isNominal s) -> A.AP s
  | C.NOT f1 -> A.NOT (exportALCFormula f1)
  | C.OR (f1, f2) -> A.OR (exportALCFormula f1, exportALCFormula f2)
  | C.AND (f1, f2) -> A.AND (exportALCFormula f1, exportALCFormula f2)
  | C.EQU (f1, f2) -> A.EQU (exportALCFormula f1, exportALCFormula f2)
  | C.IMP (f1, f2) -> A.IMP (exportALCFormula f1, exportALCFormula f2)
  | C.EX (s, f1) -> A.EX (s, false, exportALCFormula f1)
  | C.AX (s, f1) -> A.AX (s, false, exportALCFormula f1)
  | C.MIN (n, s, f1) -> A.MIN (n, s, false, exportALCFormula f1)
  | C.MAX (n, s, f1) -> A.MAX (n, s, false, exportALCFormula f1)
  | C.MORETHAN (n, s, f1) -> A.MIN (n+1, s, false, exportALCFormula f1)
  | C.MAXEXCEPT (n, s, f1) -> A.MAX (n, s, false, exportALCFormula (C.NOT f1))
  | _ -> raise (C.CoAlgException "Formula is not expressible in ALC.")

let exportALCQuery tbox (_, f) =
  let tboxND = List.map (fun (_, f) -> exportALCFormula f) tbox in
  ([exportALCFormula f], tboxND, [])


let solvCoalg sortTable args =
  G.evaluateFkt (fun (tbox, sf) -> CoAlgReasoner.isSat sortTable nomTable tbox sf) args

let solvALC (tbox, sf) =
  let args = exportALCQuery tbox sf in
  G.evaluateFkt (fun (fl, tboxND, tboxD) -> ALCGraph.isSat fl tboxND tboxD) args

let solvFact (tbox, sf) timeout =
  let (fl, tboxND, tboxD) = exportALCQuery tbox sf in
  A.createFactTBox "fact.tbox" fl tboxND tboxD;
  let fact = "./FaCT++" in
  let reg = "is satisfiable" in
  let call () = G.callExt None fact ["fact.conf"] (Str.regexp_string reg) timeout in
  let ignore_exception = false in
  if ignore_exception then try call () with _ -> (G.FAILED, -1.0, 0)
                      else call ()

let solvTatl agents (tbox,(_,sf)) timeout = (* ignore the TBox... *)
  let tatl = "./tatl" in
  let tboxlist = List.map (fun (_,f) -> CoAlgFormula.exportTatlFormula f) tbox in
  let agstr = String.concat "," (List.map string_of_int agents) in
  let tboxstr = "~<<"^agstr^">>F~ ("^(String.concat " /\\ " tboxlist) ^ ") /\\ " in
  let inp = CoAlgFormula.exportTatlFormula sf in
  let inp = if TList.empty tbox then inp else (tboxstr ^ "(" ^ inp ^ ")") in
  G.callExt (Some ("1\n"^inp^"\n4")) tatl [] (Str.regexp_string "The formula is satisfiable") timeout

let solvCool functorname (tbox,sf) timeout =
  let inp = CoAlgFormula.exportQueryParsable tbox sf in
  G.callExt (Some inp) "./coalg" ["sat"; functorname] (Str.regexp_string "1: satisfiable") timeout

let solvCoolCL = solvCool "CL"

let solvers =
  let _ = [(solvCoalg sortK, "CoAlg"); (solvALC, "alc"); (solvFact, "fact")] in
  [(solvFact, "fact"); (solvCoalg sortK, "CoAlg")]

let printUsage () =
  print_endline "Usage: \"coalgcompare [<flags>] <logic> <logicargs>\" does tests where";
  print_endline "       <flags> in { --seed INITIALSEED }";
  print_endline "       <logic> in {";
  print_endline "                     K <formula counts in tbox> <list of sizes>";
  print_endline "                     (e.g. »K 0,5,10 10000,15000,20000«)";
  print_endline "                     genCL <agents> <list of sizes>";
  print_endline "                     (e.g. »genCL 1,2,3 4000,42«)";
  print_endline "                  }";
  exit 1

let runTests solvs formulas timeout : testresults =
  let reasonerNames = List.map (fun (_,s) -> s) solvs in
  let fillCell (fn,formula) (solver,sn) =
    debugMsg ("running " ^ sn ^ " on " ^ fn);
    let (res, time, _) = solver formula timeout in
    let answer = match res with
                 | G.FINISHED (v) -> if (v) then ASAT else AUNSAT
                 | G.TIMED_OUT -> ATIMEOUT
                 | G.FAILED -> AFAILED
    in
    (answer, time)
  in
  let interrupted = ref (false) in
  let do_line (str,f) = if not !interrupted then
                            try (str,List.map (fun s -> fillCell (str,f) s) solvs)
                            with UserInterupt -> (interrupted := true; ("",[]))
                        else ("",[])
  in
  let results = List.filter (fun (_,cols) -> match cols with
                                             | (_::_) -> true
                                             | [] -> false)
                            (List.map do_line formulas)
  in
  (reasonerNames,results)


let doTestK (tboxsizes:int list) (szlist:int list) : testresults =
  let solvs = [(solvFact, "fact"); (solvCool "K", "cool")] in
  (*
    lF A list of pairs (f, n), e.g. (p, 1).
    The list must not be empty.
    lFF A list of pairs (f->f, n), e.g. (~., 1).
    The list must not be empty.
    lFFF A list of pairs (f->f->f, n), e.g. (.&., 1).
    lPFF A list of pairs (p->f->f, n), e.g. (<.>., 1).
    lP A list of pairs (p, n), e.g. (a, 1).
    The list must not be empty unless lPFF is also empty.
    lPP A list of pairs (p->p, n), e.g. ( *., 1).
    lFP A list of pairs (f->p, n), e.g. (?., 1).
    lPPP A list of pairs (p->p->p, n), e.g. (.+., 1).
  *)
  let timeout = !ptimeout in (* 5 minutes *)
  let lF = List.map (fun p -> C.AP ("p"^(string_of_int p))) (1--3) in
  let lF = List.append lF (List.map (C.const_not) lF) in
  let lF = C.TRUE :: C.FALSE :: lF in
  let lF = List.map (fun v -> (v,1)) lF in
  let roles : string list = List.map (fun p -> "R"^(string_of_int p)) (1--3) in
  let exs : (C.formula -> C.formula) list = List.map C.const_ex roles in
  let axs : (C.formula -> C.formula) list = List.map C.const_ax roles in
  let lFF = List.map (fun v -> (v,1)) (List.append exs axs) in
  let lFFF : (C.formula -> C.formula -> C.formula) list = List.append
                (List.map (fun x y z -> x (C.const_and y z)) exs)
                (List.map (fun x y z -> x (C.const_or y z))  axs)
  in
  let lFFF = List.map (fun v -> (v,1)) lFFF in
  let (genF,_) = G.mk_generator lF lFF lFFF [] [] [] [] [] in
  let s1 = 0 in (* probably the correct formula... *)
  let formulas = List.map (fun (sz,cnt) ->
                            let tbox = List.map (fun _ -> (s1,genF sz)) (1 -- cnt) in
                            ("sz"^(string_of_int sz)^"tb"^(string_of_int cnt),(tbox,(s1,genF sz)))
                          ) (TList.prod szlist tboxsizes) in
  runTests solvs formulas timeout

let doTestGenericGML () : testresults =
  let solvs = [(solvFact, "fact"); (solvCool "GML", "cool")] in
  let timeout = !ptimeout in (* 5 minutes *)
  let formulas = ref [] in
  let counter = ref 0 in
  (try
    while true do
      let input = read_line () in
      if not (GenAndComp.isEmptyString input) then
        let (tbox, f) = CoAlgFormula.importQuery input in
        incr counter;
        let test = ("line" ^ (string_of_int !counter), (tbox,f)) in
        formulas := test::(!formulas)
      else ()
    done
  with End_of_file -> ());
  let formulas = List.rev !formulas in
  runTests solvs formulas timeout


let doTestGenericK () : testresults =
  let solvs = [(solvFact, "fact"); (solvCool "K", "cool")] in
  let timeout = !ptimeout in (* 5 minutes *)
  let formulas = ref [] in
  let counter = ref 0 in
  (try
    while true do
      let input = read_line () in
      if not (GenAndComp.isEmptyString input) then
        let (tbox, f) = CoAlgFormula.importQuery input in
        incr counter;
        let test = ("line" ^ (string_of_int !counter), (tbox,f)) in
        formulas := test::(!formulas)
      else ()
    done
  with End_of_file -> ());
  let formulas = List.rev !formulas in
  runTests solvs formulas timeout



let doTestGenericCL () : testresults =
  let agents = [1;2;3;4;5] in
  let solvs = [(solvTatl agents, "tatl"); (solvCoolCL, "cool")] in
  let timeout = !ptimeout in (* 5 minutes *)
  let formulas = ref [] in
  let counter = ref 0 in
  (try
    while true do
      let input = read_line () in
      if not (GenAndComp.isEmptyString input) then
        let (tbox, f) = CoAlgFormula.importQuery input in
        incr counter;
        let test = ("line" ^ (string_of_int !counter), (tbox,f)) in
        formulas := test::(!formulas)
      else ()
    done
  with End_of_file -> ());
  let formulas = List.rev !formulas in
  runTests solvs formulas timeout



let printRawData ((rn,results):testresults) : unit =
    let doubler l = List.fold_right (fun a b -> (a ^ "_state")::(a^"_time")::b) l [] in
    print_endline (String.concat ";" ("formula"::doubler rn)) ;
    let format_cell (stat,time) = (string_of_solvanswer stat) ^ ";" ^ (string_of_float time) in
    let format_line (fn,res) = fn ^ ";" ^ (String.concat ";" (List.map format_cell res)) in
    let lines = List.map format_line results in
    print_endline (String.concat "\n" lines)

let doTestCL agents formulas : testresults =
  let solvs = [(solvTatl agents, "tatl"); (solvCoolCL, "cool")] in
  runTests solvs formulas !ptimeout

let doGenCL aglist sizelist : CoAlgFormula.formula list =
  (*
    lF A list of pairs (f, n), e.g. (p, 1).
    The list must not be empty.
    lFF A list of pairs (f->f, n), e.g. (~., 1).
    The list must not be empty.
    lFFF A list of pairs (f->f->f, n), e.g. (.&., 1).
    lPFF A list of pairs (p->f->f, n), e.g. (<.>., 1).
    lP A list of pairs (p, n), e.g. (a, 1).
    The list must not be empty unless lPFF is also empty.
    lPP A list of pairs (p->p, n), e.g. ( *., 1).
    lFP A list of pairs (f->p, n), e.g. (?., 1).
    lPPP A list of pairs (p->p->p, n), e.g. (.+., 1).
  *)
  let lF = List.map (fun p -> (C.AP ("p"^(string_of_int p)), 1)) (1--3) in
  let lFF = [(C.const_not,1)] in
  let lFFF = [(C.const_and,1)] in (* Warning: This forbids disjunctions! *)
  let lPFF = [(C.const_enforces,1); (C.const_allows, 1)] in
  (* role names *)
  let lP = List.map (fun p -> (p, 1)) (TList.powerset aglist) in
  let lPP : ((int list -> int list) * int) list= [] in
  let lFP = [] in
  let lPPP = [] in
  let (genF,_) = G.mk_generator lF lFF lFFF lPFF lP lPP lFP lPPP in
  List.map (fun sz -> (genF sz)) sizelist

let _ =
    let argidx = ref (1) in
    let fail str = (prerr_endline str ; printUsage ()) in
    let getarg () =
        if !argidx >= Array.length Sys.argv then (fail "Error: Missing argument ") else ();
        let str = Sys.argv.(!argidx) in
        argidx := 1 + !argidx;
        str
    in
    let pcmd = ref (getarg ()) in
    let rec parseflags () =
        match !pcmd with
        | "--seed" -> (pseed := (int_of_string (getarg ()));
                       pcmd := getarg ();
                       parseflags ())
        | "--timeout" -> (ptimeout := (int_of_string (getarg ()));
                       pcmd := getarg ();
                       parseflags ())
        | _ -> ()
    in
    parseflags ();
    Random.init !pseed;
    let alarmHandler (n : int) = raise G.Timeout in
    let sigintHandler (n : int) = prerr_endline "" ; raise UserInterupt in
    let oldHandler = Sys.signal Sys.sigalrm (Sys.Signal_handle alarmHandler) in
    let oldSigint = Sys.signal Sys.sigint (Sys.Signal_handle sigintHandler) in
    debugMsg ("seed is " ^ (string_of_int (!pseed)));
    let cmd = !pcmd in
    (
    match cmd with
    | "K" ->
             let tboxsizes = intlist_of_string (getarg ()) in
             let szlist = intlist_of_string (getarg()) in
             printRawData (doTestK tboxsizes szlist)
    | "genericK" ->
            printRawData (doTestGenericK ())
    | "genericGML" ->
            printRawData (doTestGenericGML ())
    | "genericCL" ->
            printRawData (doTestGenericCL ())
    | "CL1" -> let sz = (5000 -- 5080) in
               let agents = [1;2;3;4;5] in
               let ats (str,f) = (str, ([], (0, f))) in (* add tbox and sort *)
               printRawData (doTestCL agents (List.map ats (TList.zip (List.map string_of_int sz) (doGenCL agents sz))))
    | "genCL" -> let p1 = getarg () in
                 let p2 = getarg () in
                 let forms = doGenCL (intlist_of_string p1) (intlist_of_string p2) in
                 List.iter (fun f -> print_endline (CoAlgFormula.exportFormula f)) forms
    | _ -> raise (CoAlgFormula.CoAlgException ("Unknown Logic name »" ^ (Sys.argv.(1)) ^ "«"))
    );
    Sys.set_signal Sys.sigalrm oldHandler;
    Sys.set_signal Sys.sigint oldSigint


(*
let expQ (tbox, sf) = C.exportQuery tbox sf

let rateF = int_of_string Sys.argv.(1)
let rateP = int_of_string Sys.argv.(2)
let nr = int_of_string Sys.argv.(3)
let start = int_of_string Sys.argv.(4)
let stop = int_of_string Sys.argv.(5)
let step = int_of_string Sys.argv.(6)

let litFkt n = n / rateF
let pFkt n = n / rateP
let tbcardFkt n = 0
let tbfsizeFkt n = max 1 (n / 10)

let genQ = genQuery litFkt pFkt tbcardFkt tbfsizeFkt

let _ = G.toGnuPlot ~dots:10 ~timeout:10 ~fail:true expQ genQ solvers nr start stop ((+) step)

*)

(*
let files = [ ("benchmarks/DL98/DL98_alc_branch_sat.txt", "branch\\_n");
              ("benchmarks/DL98/DL98_alc_branch_unsat.txt", "branch\\_p");
              ("benchmarks/DL98/DL98_alc_d4_sat.txt", "d4\\_n");
              ("benchmarks/DL98/DL98_alc_d4_unsat.txt", "d4\\_p");
              ("benchmarks/DL98/DL98_alc_dum_sat.txt", "dum\\_n");
              ("benchmarks/DL98/DL98_alc_dum_unsat.txt", "dum\\_p");
              ("benchmarks/DL98/DL98_alc_grz_sat.txt", "grz\\_n");
              ("benchmarks/DL98/DL98_alc_grz_unsat.txt", "grz\\_p");
              ("benchmarks/DL98/DL98_alc_lin_sat.txt", "lin\\_n");
              ("benchmarks/DL98/DL98_alc_lin_unsat.txt", "lin\\_p");
              ("benchmarks/DL98/DL98_alc_path_sat.txt", "path\\_n");
              ("benchmarks/DL98/DL98_alc_path_unsat.txt", "path\\_p");
              ("benchmarks/DL98/DL98_alc_ph_sat.txt", "ph\\_n");
              ("benchmarks/DL98/DL98_alc_ph_unsat.txt", "ph\\_p");
              ("benchmarks/DL98/DL98_alc_poly_sat.txt", "poly\\_n");
              ("benchmarks/DL98/DL98_alc_poly_unsat.txt", "poly\\_p");
              ("benchmarks/DL98/DL98_alc_t4p_sat.txt", "t4p\\_n");
              ("benchmarks/DL98/DL98_alc_t4p_unsat.txt", "t4p\\_p") ]

let prebenchmarks =
  let mapB (filename, lab) =
    let file = open_in filename in
    let res = ref [] in
    let _ =
      try
        while true do
          let input = input_line file in
          if not (G.isEmptyString input) then
            let q = C.importQuery input in
            res := q::!res
          else ()
        done
      with End_of_file -> ()
    in
    (List.rev !res, lab)
  in
  List.map mapB files

let benchmarks =
  List.map (fun f -> fun () -> f) (List.flatten (List.map (fun (fl, _) -> fl) prebenchmarks))

let labels = List.map (fun (fl, lab) -> (List.length fl, lab)) prebenchmarks

let timeouts = [ 1; 2; 4; 8; 15; 30; 60 ]

let _ = G.outputBenchmarks ~fail:true expQ solvers timeouts benchmarks labels
 *)

let logicFkts = {
  G.l_tt = C.TRUE;
  G.l_ff = C.FALSE;
  G.l_ap = (fun s i -> C.AP (s ^ (string_of_int i)));
  G.l_not = C.const_not;
  G.l_and = C.const_and;
  G.l_or = C.const_or;
  G.l_imp = C.const_imp;
  G.l_aa = (fun s i -> s ^ (string_of_int i));
  G.l_box = C.const_ax;
  G.l_dia = C.const_ex;
  G.l_aw = (fun _ _ -> raise (C.CoAlgException "not implemented"));
  G.l_ev = (fun _ _ -> raise (C.CoAlgException "not implemented"));
  G.l_until = (fun _ _ -> raise (C.CoAlgException "not implemented"))
}

(*
let m = int_of_string Sys.argv.(1)
let k = int_of_string Sys.argv.(2)

let wrapSorts (f, tbox) =
  let tbox1 = List.map (fun f -> (0, f)) tbox in
  (tbox1, (0, f))

let (tbox, sf) =
  let (concl, tbox) = G.mk_wisemen_tbox logicFkts m k in
  wrapSorts (C.NOT concl, tbox)

let (tbox, sf) =
  let (concl, tbox) = G.mk_muddychildren_tbox logicFkts m k in
  wrapSorts (C.NOT concl, tbox)

let (tbox, sf) = wrapSorts (G.mk_exptime_tbox logicFkts m false)

let _ = CoAlgReasoner.isSat ~verbose:true sorts nomTable tbox sf
let _ =
  let (fl, tboxND, tboxD) = exportALCQuery tbox sf in
  ALCGraph.isSat ~verbose:true fl tboxND tboxD
 *)

(* vim: set et sw=2 sts=2 ts=8 : *)
