(** Accepts formulae from the standard input and tests them for satisfiability.
    Each formula has to be on a separate line and empty lines are ignored.
    The input is terminated by EOF.
    @author Florian Widmann
 *)

module CM = CoAlgMisc
module CF = CoAlgFormula

module FE = FunctorParsing
module Reasoner = CoAlgReasoner

open CoolUtils

(* A partial function mapping nominals to their sorts. *)
let nomTable name =
  assert (CoAlgFormula.isNominal name);
  Some 0


let _ = Gc.set { (Gc.get()) with Gc.minor_heap_size = 4194304; major_heap_increment = 67108864; space_overhead = 120 }

let printUsage () =
  print_endline "Usage: \"debugger <functor> <formula>\" where";
  print_endline "";
  print_endline "Examples:";
  print_endline "./debugger.native  '(Const a b) * (Const c) * (Const d)'  '( [pi1] =a <=> [pi2][pi1] =c) & ([pi1] =b <=> [pi2][pi2] =d)'";
  exit 1

let printResult () =
    let str = match Reasoner.isRootSat () with
    | None -> "Reasoning not finished yet"
    | Some sat -> "Reasoning finished: "^(if sat then "Sat" else "Unsat")
    in
    print_endline str

let printStatus () =
    print_endline ("Generated states: " ^ (string_of_int (CM.graphSizeState ())));
    print_endline ("Generated cores: " ^ (string_of_int (CM.graphSizeCore ())));
    print_endline ("Generated constraints: " ^ (string_of_int (CM.graphSizeCnstr ())));
    print_endline "";
    print_endline "Queue content:";
    print_endline (CM.queuePrettyStatus ());
    printResult ()

let rec repeat (n:int) (action: unit -> unit) =
    if n <= 0
    then ()
    else (action() ; repeat (n-1) action)

let doStep args =
    let action() =
        if Reasoner.reasonerFinished ()
        then print_endline "Reasoner already finished!"
        else try
                Reasoner.runReasonerStep ()
             with CoAlgMisc.CoAlg_finished _ -> printResult()
    in
    let steps = match args with
                | (str::_) -> (int_of_string str)
                | [] -> 1
    in
    try repeat steps action
    with Reasoner.ReasonerError str -> print_endline ("Error: "^str)

let listCores args =
    CM.graphIterCores (fun core -> print_endline (CM.coreToString core))

let listStates args =
    CM.graphIterStates (fun state -> print_endline (CM.stateToString state))

let exportGraph args =
  print_endline "digraph reasonerstate {";
  CM.graphIterCores (fun core -> print_endline (CM.coreToDot core));
  CM.graphIterStates (fun state -> print_endline (CM.stateToDot state));
  print_endline "}"

let showNode = function
    | (n::_) ->
        let n = int_of_string n in

        CM.graphIterCores (fun core ->
            if CM.coreGetIdx core = n
            then print_endline (CM.coreToString core)
            else ());
        CM.graphIterStates (fun state ->
            if CM.stateGetIdx state = n
            then print_endline (CM.stateToString state)
            else ())
    | [] -> print_endline "Error: number required!"

let _ =
  if Array.length Sys.argv < 2 then printUsage()
  else
    let formula = Sys.argv.(2) in
    let sorts = (FE.sortTableFromString Sys.argv.(1)) in
    let (tbox, f) = CoAlgFormula.importQuery formula in
    Reasoner.initReasoner sorts nomTable tbox f;
    let session = {
      Repl.prompt = "> ";
      Repl.bindings = Repl.addHelpBinding [
        Repl.bind "s" doStep "Runs another reasoner step"
            "step [n] runs n reasoner steps, where n is optional and defaults to 1"
        ;
        Repl.bind "step" doStep "Runs another reasoner step" "" ;
        Repl.bind "st" (fun _ -> printStatus ()) "Prints reasoning status" "";
        Repl.bind "status" (fun _ -> printStatus ()) "Prints reasoning status" "";
        Repl.bind "lscores" listCores "Lists all cores" "";
        Repl.bind "lsstates" listStates "Lists all states" "";
        Repl.bind "graphviz" exportGraph "Graphviz source of reasoner state" "";
        Repl.bind "node" showNode "shows details of a node" "usage: node n";
        Repl.exitBinding
      ];
    }
    in
    Repl.start session

(* vim: set et sw=2 sts=2 ts=8 : *)
