
module L = List

open CoolUtils
open Str
open Stream


(* types *)
type annotation = string * int * int (* file and line number and column*)

exception ParseError of annotation * string
exception OWLParseError of string

type 'a tree =
    | Leaf of 'a
    | Node of (('a tree) list)

type 'a intree = (* tree carrying values in leafes and inner nodes *)
    | INode of 'a * ('a intree list)

type 'a annotated = 'a * annotation

(* implementation *)

let stream_map f stream =
    let next i =
      try Some (f (Stream.next stream))
      with Stream.Failure -> None in
    Stream.from next

let stream_fold f stream init =
    let result = ref init in
    Stream.iter
      (fun x -> result := f x !result)
      stream;
    !result;;

let stream_combine stream1 stream2 =
    let rec next i =
      try Some (Stream.next stream1, Stream.next stream2)
      with Stream.Failure -> None in
    Stream.from next;;

let line_stream_of_string string =
    Stream.of_list (Str.split (Str.regexp "\n") string);;

let list_of_stream stream =
    let cons a b = a::b in
    L.rev (stream_fold cons stream [])

let rec intree_flatten (INode (v,sub)) : 'a list =
    v :: L.concat (L.map intree_flatten sub)

let tokens_of_string_stream (lines:string Stream.t) : string annotated list =
    let lastline = ref 0 in
    let nextline () =
        lastline := 1 + !lastline;
        !lastline
    in
    let tokenize l =
        let col = ref 1 in
        let linenr = nextline () in
        let annotation s =
            let c = !col in
            col := !col + (String.length s);
            ("",linenr, c)
        in
        let extract x = match x with
                        | (Delim a) -> a
                        | (Text  a) -> a
        in
        let words = L.map extract (Str.full_split (regexp "[ \t+]") l) in
        let tokens = L.concat (L.map (full_split (regexp "[()]")) words) in
        L.map (fun t -> (t, annotation t)) (L.map extract tokens)
    in
    L.concat (list_of_stream (stream_map tokenize lines))

let string_of_annotation (file,line,col) =
    file ^ ":" ^ (string_of_int line) ^ ":" ^ (string_of_int col)

let string_of_annotated str_of_a (obj,anno) : string =
    string_of_annotation anno ^ " " ^ str_of_a obj

let filter_comment line =
    let comment_reg = "//.*$" in
    let lst = full_split (regexp "<[^>]*>") line in
    let lastidx = (List.length lst) - 1 in
    let f i a =
        match a with
        | Delim a -> a
        | Text a -> if not (i = lastidx) then a
                    else replace_first (regexp comment_reg) "" a
    in
    String.concat "" (L.mapi f lst)


let tree_of_tokens lst : string annotated tree =
    let stream:string annotated Stream.t = Stream.of_list lst in
    (* returns the next tree in our forest *)
    let rec tree_of_token_stream _ : string annotated tree option =
        match peek stream with
        | Some ("(", a1) ->
            junk stream;
            (* assemble another Node by creating children to that node *)
            (* until a ")" appears *)
            let composenode _ : string annotated tree option = 
                match peek stream with
                | Some (")", _) -> (junk stream ; None)
                | Some (_,_) -> tree_of_token_stream ()
                | None -> raise (ParseError (a1, "Unmatched ("))
            in
            Some (Node (list_of_stream (Stream.from composenode)))
        | Some annostr -> junk stream ; Some (Leaf annostr)
        | None -> None
    in
    let forest = list_of_stream (Stream.from (fun _ -> tree_of_token_stream ())) in
    Node forest

let string_of_tree str_of_a atree : string =
    let rec lines_of_tree t = 
        let firstchild = ref true in
        let indent str =
            if !firstchild
            then (firstchild := false; "---+"^str)
            else  "   |"^str
        in
        match t with
        | (Leaf a) -> ["-" ^ str_of_a a]
        | (Node lst) -> (L.map indent (L.concat (L.map lines_of_tree lst)))
    in
    String.concat "\n" (lines_of_tree atree)

let string_of_intree str_of_a atree : string =
    let rec lines_of_tree (INode (a,lst)) = 
        let indent str = "  |"^str in
        ("--+ "^str_of_a a)::
        (L.map indent (L.concat (L.map lines_of_tree lst)))
    in
    String.concat "\n" (lines_of_tree atree)


let filter_whitespace =
    (* only keep those strings that aren't whitespace only *)
    L.filter (fun (s,_) -> not (string_match (regexp "[ ]*$") s 0))

let tree_of_string str =
    let lns = stream_map filter_comment (line_stream_of_string str) in
    tree_of_tokens (filter_whitespace (tokens_of_string_stream lns))

let ensure_node (t: string annotated tree) : string annotated tree list =
    match t with
    | (Leaf (s,a)) -> raise (ParseError (a, "Expected \"(\" but got "^s))
    | (Node lst) -> lst

let rec intree_list_of_tree tree =
    let nextINode (stream: 'a tree Stream.t): ('a intree option) =
        match peek stream with
        | Some (Node lst) -> None
        | Some (Leaf a) ->
            junk stream;
            let lst = match peek stream with
                | None
                | Some (Leaf _) -> []
                | Some (Node lst) -> (
                    junk stream;
                    intree_list_of_tree (Node lst)
                )
            in Some (INode (a,lst))
        | None -> None
    in
    match tree with
    | (Leaf a) -> [INode (a,[])]
    | (Node lst) ->
        let rec intree_list_of_tree_stream stream =
            match nextINode stream with
            | Some n -> (n::(intree_list_of_tree_stream stream))
            | None -> []
        in
        intree_list_of_tree_stream (Stream.of_list lst)

let rec tree_of_intree it =
    let tree_list_of_intree (INode (x,lst)) =
        (Leaf x)::(L.map tree_of_intree lst)
    in
    Node (tree_list_of_intree it)

let parse_uri (str:string) : OWL.uri =
    Str.replace_first (Str.regexp "<\\(.*\\)>") "\\1" str

let parse_prefix a lst : OWL.prefix =
    match lst with
    | [INode ((str,a),[])] -> (
        match split (regexp "=") str with
        | [l;r] -> (l,parse_uri r)
        | _ -> raise (ParseError (a,("Invalid Prefix-Definition")))
        )
    | _ -> raise (ParseError (a,("Missing Prefix-definition")))

let rec parse_role (INode ((str,a),sub)) : OWL.role =
    if ((String.contains str ':') && (sub = [])) then OWL.ROLE str
    else match str, sub with
    | "ObjectInverseOf",[r] -> OWL.INVERSE (parse_role r)
    | _,_ -> raise (ParseError (a, "Unknown role construct \""^str^"\""))

let rec parse_class_exp (INode ((name,anno), sub): string annotated intree) : OWL.class_exp =
    let pr = parse_role in
    let pc = parse_class_exp in
    if (String.contains name ':') then
        match name with
        | "owl:Thing" -> OWL.THING
        | "owl:Nothing" -> OWL.NOTHING
        | _ -> (OWL.NAMED name)
    else match name, sub with
         | "ObjectSomeValuesFrom", [role; obj] -> OWL.SOME (pr role, pc obj)
         | "ObjectAllValuesFrom", [role; obj] -> OWL.ALL (pr role, pc obj)
         | "ObjectIntersectionOf", _ -> OWL.AND (L.map pc sub)
         | "ObjectUnionOf", _ -> OWL.OR (L.map pc sub)
         | "ObjectComplementOf", [c] -> OWL.NOT (pc c)
         | "ObjectMinCardinality", _
         | "ObjectMaxCardinality", _
         | "ObjectExactCardinality", _
         | "ObjectOneOf", _
         | "ObjectHasValue", _
         | "ObjectHasSelf", _
         | "DataSomeValuesFrom", _
         | "DataAllValuesFrom", _
         | "DataHasValue", _
         | "DataMinCardinality", _
         | "DataMaxCardinality", _
         | "DataExactCardinality", _ -> raise (ParseError (anno, name^" not implemented yet."))
         | _,_ ->
           raise (ParseError(anno, "Unknown class expression name \""^name^"\" or wrong parameter count"))

let parse_axiom (ax: string annotated intree) : OWL.axiom =
    let INode ((name,anno), sub) = ax in
    let pce = parse_class_exp in
    match name with
    | "Annotation" -> let sub = L.map intree_flatten sub in (* flatten structure *)
                      let sub = L.concat sub in             (* flatten more...*)
                      let sub = L.map (fun (s,_) -> s) sub in (* drop annotations *)
                      OWL.ANNOTATION (String.concat " " sub)
    | "Declaration" -> (match sub with
                        | [(INode ((name,anno),[INode ((child,_),[])]))] ->
                            (match name with
                             | "Class" -> OWL.DECLARATION (OWL.CLASS child)
                             | _ -> raise (ParseError (anno, "Unknown declaration name \""^name^"\"")))
                        | _ ->  raise (ParseError (anno, name^" expects exactly one parameters")))
    | "SubClassOf"  -> (match sub with
                        | [csub;csuper] -> OWL.SUBCLASS (pce csub,parse_class_exp csuper)
                        | _ -> raise (ParseError (anno, name^" expects exactly two parameters")))
    | "EquivalentClasses" -> OWL.EQUIVALENTCLASSES (L.map pce sub)
    | "DisjointClasses" -> OWL.DISJOINTCLASSES (L.map pce sub)
    | "DisjointUnion" -> (match sub with
                          | (union::parts) -> OWL.DISJOINTUNION ((pce union),(L.map pce parts))
                          | _ -> raise (ParseError (anno, name^" expects at least one parameter")))
    | _ -> raise (ParseError (anno, "Unknown axiom name \""^name^"\""))

let parse_ontology a (axioms:string annotated intree list) : OWL.ontology =
    (* the first element of axioms, i.e. of the children of the ontology node
    is just the ontology name/uri *)
    match axioms with
    | (INode ((ontoname,_), [])::axioms) -> (parse_uri ontoname, L.map parse_axiom axioms)
    | _ -> raise (ParseError (a,("Missing Ontology name")))

let ap (lref:'a list ref) (el:'a): unit = (* append to 'a list ref *)
        (lref := ((!lref) @ [el]))

let parse_intree_list (strAnIts: string annotated intree list) =
    let prefs : OWL.prefixes ref = ref [] in
    let ontos : OWL.ontology list ref = ref [] in
    foreach_l strAnIts (fun f -> match f with
    | INode (("Prefix",   a), lst) -> (ap prefs (parse_prefix a lst))
    | INode (("Ontology", a), axioms) -> (ap ontos (parse_ontology a axioms))
    | INode ((n, a), _) -> raise (ParseError (a,(
        "Unknown toplevel element \""^n^"\""))));
    (!prefs, !ontos)

let parse str =
    try (let tree = tree_of_string str in
         let itlist = intree_list_of_tree tree in
         parse_intree_list itlist
    ) with ParseError ((file,line,col), msg) -> ( (* beautify error message *)
        let stream = line_stream_of_string str in
        repeat (line - 1) (fun () -> Stream.junk stream);
        let content = Stream.next stream in
        let indent = ref "" in
        repeat (col - 1) (fun () -> indent := " " ^ !indent);
        let msg = msg ^"\n"^ content ^"\n"^ !indent ^"^-- here" in
        raise (ParseError((file,line,col), msg))
    )


(* vim: set et sw=2 sts=2 ts=8 : *)
